"""
Portions of this module require that the testcontainers package is installed, as described below in the docstrings of
this module's functions and classes.
"""
from contextlib import ExitStack
from copy import deepcopy, copy
from typing import Optional, Dict, List, Type, Callable, Mapping, MutableMapping, NamedTuple
from aiohttp import web
from heaobject.registry import Resource
from heaobject.root import DesktopObjectDict
from .. import runner
from .. import wstl
from ..db.database import CollectionKey, MicroserviceDatabaseManager
from ..db.database import get_collection_key_from_name
from .mockmongo import MockMongoManager
from ..util import TypeEnforcingFrozenDataclass
from dataclasses import dataclass
from contextlib import contextmanager, closing
from collections.abc import Generator, Sequence, MutableSequence
from abc import ABC
from ..util import retry
import logging


@dataclass(frozen=True)
class DockerVolumeMapping(TypeEnforcingFrozenDataclass):
    """
    Docker volume mapping. This class is immutable.

    Attributes:
        host: the host file or directory to map.
        container: the container directory.
        mode: the read-write mode (default 'ro' for read-only).
    """
    host: str
    container: str
    mode: str = 'ro'


@dataclass(frozen=True)
class BindPort(TypeEnforcingFrozenDataclass):
    """
    Bind port pair.

    Attributes:
        host: the host port.
        container: the container port to which the host port is mapped.
    """
    host: int
    container: int

    def pair(self) -> str:
        """
        Returns a host:container port mapping string.
        """
        return f'{self.host}:{self.container}'

    def __str__(self):
        """
        Returns a host:container port mapping string.
        """
        return self.pair()

    @staticmethod
    def from_port_mapping_string(pair: str) -> 'BindPort':
        """
        Factory for creating a bind port from a host:container port mapping string.
        """
        host_, container_ = pair.split(':')
        return BindPort(int(host_), int(container_))

class DockerContainerConfig:
    """
    Docker image and configuration for starting a container. This class is immutable.
    """

    def __init__(self, image: str, ports: list[int] | None, check_path: Optional[str] = None,
                 check_port: int | None = None, check_status: int = 200,
                 volumes: Optional[List[DockerVolumeMapping]] = None, env_vars: Optional[Mapping[str, str]] = None,
                 bind_ports: list[BindPort] | None = None):
        """
        Constructor.

        :param image: the image tag (required).
        :param ports: the exposed network ports.
        :param check_path: the URL path to check if the microservice is running.
        :param check_port: the network port to use with the check_path. If None or omitted, the first port in the ports
        list will be used.
        :param check_status: the status code expected from the check_path when successful.
        :param volumes: a list of volume mappings.
        :param env_vars: a dict containing environment variable names mapped to string values.
        :param bind_ports: the bind ports for the container as two tuples (host and container).
        """
        if image is None:
            raise ValueError('image cannot be None')
        if any(not isinstance(port, int) for port in ports or []):
            raise ValueError('ports must contain only ints')
        if any(not isinstance(volume, DockerVolumeMapping) for volume in volumes or []):
            raise TypeError(f'volumes must contain only {DockerVolumeMapping} objects')
        if any(not isinstance(k, str) and isinstance(v, str) for k, v in (env_vars or {}).items()):
            raise TypeError('env_vars must be a str->str dict')
        if any(not isinstance(bind_port, BindPort) for bind_port in bind_ports or []):
            raise TypeError('bind_ports must contain BindPort objects')
        self.__image = str(image)
        self.__ports = list(ports) if ports is not None else []
        self.__bind_ports = list(bind_ports) if bind_ports is not None else []
        self.__check_path = str(check_path) if check_path is not None else None
        self.__check_port = int(check_port) if check_port is not None else self.__ports[0]
        self.__check_status = int(check_status)
        self.__volumes = list(volumes) if volumes else []
        self.__env_vars = dict(env_vars) if env_vars is not None else {}

    @property
    def image(self) -> str:
        """
        The image tag (read-only).
        """
        return self.__image

    @property
    def ports(self) -> list[int]:
        """
        The exposed ports (read-only).
        """
        return list(self.__ports)

    @property
    def bind_ports(self) -> list[BindPort]:
        return list(self.__bind_ports)

    @property
    def check_path(self) -> Optional[str]:
        """
        The URL path to check for whether the microservice is running (read-only).
        """
        return self.__check_path

    @property
    def check_port(self) -> int:
        """
        The network port to use with the check_path. If None or omitted, the first port in the ports
        list will be used.
        """
        return self.__check_port

    @property
    def check_status(self) -> int:
        """
        The status code expected from a successful call to the check path. Is 200 by default.
        """
        return self.__check_status

    @property
    def volumes(self) -> List[DockerVolumeMapping]:
        """
        A list of VolumeMapping instances indicating what volumes to map (read-only, never None).
        """
        return copy(self.__volumes)

    @property
    def env_vars(self) -> Dict[str, str]:
        """
        A dict of environment variable names to string values.
        """
        return copy(self.__env_vars)

    def with_env_vars(self, env_vars: Optional[Mapping[str, str]]) -> 'DockerContainerConfig':
        """
        Returns a new DockerContainerConfig with the same values as this one, plus any environment variables in the
        env_vars argument.

        :param env_vars: any environment variables.
        :return:
        """
        new_env_vars = self.env_vars
        if env_vars is not None:
            new_env_vars.update(env_vars)
        return DockerContainerConfig(image=self.image, ports=self.ports, check_path=self.check_path,
                                     volumes=self.volumes, env_vars=new_env_vars)

    def get_bridge_components(self, bridge_url: str) -> list[DesktopObjectDict]:
        return []

    def get_external_components(self, external_url: str) -> list[DesktopObjectDict]:
        return []


class MicroserviceContainerConfig(DockerContainerConfig):
    """
    Docker image and configuration for starting a microservice. This class is immutable.
    """

    def __init__(self, image: str, port: int, check_path: str | None = None, check_status: int = 200,
                 resources: List[Resource] | None = None,
                 volumes: List[DockerVolumeMapping] | None = None,
                 env_vars: Mapping[str, str] | None = None,
                 db_manager_cls: type[MicroserviceDatabaseManager] | None = None):
        """
        Constructor.

        :param image: the image tag (required).
        :param port: the exposed port (required).
        :param check_path: the URL path to check if the microservice is running.
        :param check_status: the status code expected from the check_path when successful.
        :param resources: a list of heaobject.registry.Resource dicts indicating what content types this image is designed for.
        :param volumes: a list of volume mappings.
        :param env_vars: a dict containing environment variable names mapped to string values.
        :param db_manager_cls: the database manager type for this microservice, or None if the service does not use a
        database.
        """
        if port is None:
            raise ValueError('port cannot be None')
        super().__init__(image=image, ports=[port], check_path=check_path, check_status=check_status, volumes=volumes,
                         env_vars=env_vars)
        if any(not isinstance(r, Resource) for r in resources or []):
            raise TypeError(f'resources must contain only {Resource} objects')
        self.__port = int(port)
        self.__resources = [deepcopy(e) for e in resources or []]
        self.__volumes = list(volumes) if volumes else []
        self.__env_vars = dict(env_vars) if env_vars is not None else {}
        self.__db_manager_cls = db_manager_cls  # immutable

    @property
    def port(self) -> int:
        """
        The exposed port (read-only).
        """
        return self.ports[0]

    @property
    def resources(self) -> Optional[List[Resource]]:
        """
        A list of heaobject.registry.Resource dicts indicating what content types this image is designed for (read-only).
        """
        return deepcopy(self.__resources)

    @property
    def env_vars(self) -> Dict[str, str]:
        """
        A dict of environment variable names to string values.
        """
        return copy(self.__env_vars)

    @property
    def db_manager_cls(self) -> Optional[type[MicroserviceDatabaseManager]]:
        """
        The database manager class, if any.
        """
        return self.__db_manager_cls  # immutable

    def with_env_vars(self, env_vars: Optional[Mapping[str, str]]) -> 'MicroserviceContainerConfig':
        """
        Returns a new MicroserviceContainerConfig with the same values as this one, plus any environment variables in
        the env_vars argument.

        :param env_vars: any environment variables.
        :return:
        """
        new_env_vars = self.env_vars
        if env_vars is not None:
            new_env_vars.update(env_vars)
        return MicroserviceContainerConfig(image=self.image, port=self.port, check_path=self.check_path,
                                           resources=self.resources, volumes=self.volumes, env_vars=new_env_vars,
                                           db_manager_cls=self.db_manager_cls)

    def get_bridge_components(self, bridge_url: str) -> list[DesktopObjectDict]:
        logger = logging.getLogger(__name__)
        result = super().get_bridge_components(bridge_url)
        logger.debug('Adding bridge component for resources %s', self.resources)
        result.append({'type': 'heaobject.registry.Component', 'base_url': bridge_url, 'name': bridge_url,
                       "owner": "system|none",
                       'resources': [r.to_dict() for r in self.resources or []]})
        return result

    def get_external_components(self, external_url: str) -> list[DesktopObjectDict]:
        logger = logging.getLogger(__name__)
        result = super().get_external_components(external_url)
        logger.debug('Adding external component for resources %s', self.resources)
        result.append({'type': 'heaobject.registry.Component', 'base_url': external_url, 'name': external_url,
                       "owner": "system|none", 'resources': [r.to_dict() for r in self.resources or []]})
        return result


class RegistryContainerConfig(MicroserviceContainerConfig, ABC):
    """
    Abstract base class for builders that configure and create HEA Registry Service docker containers.

    This class assumes that the testcontainers package is installed. Do not create instances of it when testcontainers
    will not be available, for example, in any code that needs to run outside automated testing or the SwaggerUI
    interface. Using it as a type annotation for optional parameters and the like where no actual instances of it
    will be created is okay, however.
    """

    def __init__(self, *args, db_manager_cls: type[MicroserviceDatabaseManager], **kwargs):
        super().__init__(*args, **kwargs)
        self.__db_manager_cls = db_manager_cls

    @property
    def db_manager_cls(self) -> type[MicroserviceDatabaseManager]:
        return self.__db_manager_cls  # type:ignore  # mypy bug? It won't accept removing Optional from the return type.


class MockRegistryContainerConfig(RegistryContainerConfig):
    """
    Creates an HEA Registry Service docker container configured to use a mock mongodb database for data management.

    This class assumes that the testcontainers package is installed. Do not create instances of it when testcontainers
    will not be available, for example, in any code that needs to run outside automated testing or the SwaggerUI
    interface. Using it as a type annotation for optional parameters and the like where no actual instances of it
    will be created is okay, however.

    :param image: the label of the docker image to use (required).
    """

    def __init__(self, image: str):
        super().__init__(image=image, port=8080, check_path='/components', db_manager_cls=MockMongoManager)


class DockerContainerPorts(NamedTuple):
    """
    A docker image and its external and bridge URLs.

    Parameters:
        image: the name of the image.
        external_url: the URL of the container from outside the docker bridge network as a string.
        bridge_url: the URL of the container on the docker bridge network as a string.
    """
    image: str
    external_url: str
    bridge_url: str


@contextmanager
def app_context(db_manager_cls: Type[MicroserviceDatabaseManager],
                desktop_objects: Mapping[CollectionKey, Sequence[DesktopObjectDict]],
                other_docker_images: Optional[Sequence[DockerContainerConfig]] = None,
                registry_docker_image: Optional[RegistryContainerConfig] = None,
                content: Mapping[CollectionKey, Mapping[str, bytes]] | None = None,
                wstl_builder_factory: Optional[Callable[[], wstl.RuntimeWeSTLDocumentBuilder]] = None,
                package_name: str | None = None) -> Generator[
    tuple[web.Application, list[DockerContainerPorts]], None, None]:
    """
    Starts the test environment. The test environment consists of: a "bridge" database that is accessible from the
    internal docker network; an "external" database that is accessible from outside the network; a "bridge" registry
    service that is accessible from the internal docker network; an "external" registry service that is accessible from
    outside the network; the service being tested, which is run from outside of docker; and any service dependencies,
    which are run as docker containers. The provided context manager will clean up any resources upon exit.

    Do not pass DockerContainerConfig nor RegistryContainerConfig objects into this function unless the testcontainer
    package is installed.


    :param db_manager_cls: the database manager class for the microservice being tested (required).
    :param desktop_objects: HEA desktop objects to load into the database (required), as a map of collection -> list of
    desktop object dicts.
    :param other_docker_images: the docker images of any service dependencies.
    dictionary keys are strings. If None, defaults to db_manager_cls.
    :param registry_docker_image: an HEA registry service docker image.
    :param content: any content to load into the database.
    :param wstl_builder_factory: a zero-argument callable that will return a RuntimeWeSTLDocumentBuilder. Optional if
    this service has no actions. Typically, you will use the heaserver.service.wstl.get_builder_factory function to
    get a factory object.
    :param package_name: the service's distribution package name, used to dynamically register metadata about this
    service with the registry service. Do not pass a package name if no registry service is running.
    :return: a generator of two tuples containing the aiohttp Application object and a list of DockerContainerPorts
    containing the ports for each requested docker image.
    """

    bridge_desktop_objects: MutableMapping[CollectionKey, MutableSequence[DesktopObjectDict]] = dict({k: list(v) for k, v in deepcopy(desktop_objects).items()})
    external_desktop_objects: MutableMapping[CollectionKey, MutableSequence[DesktopObjectDict]] = dict({k: list(v) for k, v in deepcopy(desktop_objects).items()})

    def _bridge_dbs_to_start() -> set[type[MicroserviceDatabaseManager]]:
        bridge_db_manager_cls = set()
        if other_docker_images:
            bridge_db_manager_cls.update(
                [img.db_manager_cls for img in other_docker_images if
                 isinstance(img, MicroserviceContainerConfig) and img.db_manager_cls is not None])
        if registry_docker_image is not None and registry_docker_image.db_manager_cls is not None:
            bridge_db_manager_cls.add(registry_docker_image.db_manager_cls)
        return bridge_db_manager_cls

    class TestEnvExitStack(ExitStack):
        @retry(ValueError, IOError, retries=6, cooldown=10)
        def __exit__(self, exc_type, exc_val, exc_tb) -> bool | None:
            return super().__exit__(exc_type, exc_val, exc_tb)

    with TestEnvExitStack() as context_manager, closing(db_manager_cls()) as external_db_, \
        db_manager_cls.environment(), db_manager_cls.context():

        bridge_dbs = [context_manager.enter_context(closing(bridge_db_cls())) for bridge_db_cls in
                        _bridge_dbs_to_start()]

        external_db_.start_database(context_manager)
        for bridge_db in bridge_dbs:
            bridge_db.start_database(context_manager)

        if registry_docker_image is not None:
            (_, bridge_registry_url), (external_registry_url, _) = map(
                lambda dbs: _start_registry_service(dbs, context_manager, registry_docker_image), [bridge_dbs, [external_db_]])
        else:
            external_registry_url = None
            bridge_registry_url = None

        if other_docker_images:
            if registry_docker_image is None:
                raise ValueError('registry_docker_image is required when other_docker_images is non-None')
            docker_image_ports = _start_other_docker_containers(
                bridge_desktop_objects,
                external_desktop_objects,
                other_docker_images, bridge_registry_url,
                bridge_dbs,
                context_manager,
                registry_docker_image.db_manager_cls)
        else:
            docker_image_ports = []
        if registry_docker_image is not None:
            docker_image_ports.append(DockerContainerPorts(image=registry_docker_image.image,
                                                            external_url=external_registry_url,
                                                            bridge_url=bridge_registry_url))

        config_file = _generate_config_file(external_db_, external_registry_url)
        config_ = runner.init(config_string=config_file)
        external_db_.insert_all(external_desktop_objects, content)
        for b_ in bridge_dbs:
            b_.insert_all(bridge_desktop_objects, content)

        yield runner.get_application(package_name=package_name, db=external_db_, wstl_builder_factory=wstl_builder_factory,
                                     config=config_), docker_image_ports


def _start_registry_service(dbs, context_manager, registry_docker_image):
    from .docker import start_microservice_container
    bridge_config_ = _add_db_config(registry_docker_image, dbs)
    _, bridge_registry_url = start_microservice_container(bridge_config_, context_manager)
    return _, bridge_registry_url


def _add_db_config(docker_container_config: MicroserviceContainerConfig,
                   dbs: list[MicroserviceDatabaseManager]) -> MicroserviceContainerConfig:
    """
    Returns a copy of the docker_container_config with additional environment variables needed for connecting to the
    database.

    This function assumes that the testcontainers package is installed. Do not use it when testcontainers
    will not be available, for example, in any code that needs to run outside automated testing or the SwaggerUI
    interface.

    :param docker_container_config: a DockerContainerConfig (required).
    :param dbs: the available database containers.
    :return: a newly created DockerContainerConfig.
    """
    db_manager = _get_db_manager(dbs, docker_container_config.db_manager_cls)
    if db_manager is not None:
        env_vars = db_manager.get_env_vars()
    else:
        env_vars = None
    return docker_container_config.with_env_vars(env_vars)


def _get_db_manager(dbs: list[MicroserviceDatabaseManager],
                    db_manager_cls_: Optional[type[MicroserviceDatabaseManager]]) -> Optional[
    MicroserviceDatabaseManager]:
    """
    Returns the database manager with the given type.

    :param dbs: the available database managers.
    :param db_manager_cls_: the type of interest.
    :return: a database manager, or None if no database manager with the given type is available.
    """
    if db_manager_cls_ is not None:
        return next((b for b in dbs if isinstance(b, db_manager_cls_)), None)
    else:
        return None


def _start_other_docker_containers(
    bridge_desktop_objects: MutableMapping[CollectionKey, MutableSequence[DesktopObjectDict]],
    external_desktop_objects: MutableMapping[CollectionKey, MutableSequence[DesktopObjectDict]],
    other_docker_images: Optional[Sequence[DockerContainerConfig]],
    registry_url: Optional[str],
    bridge_dbs: list[MicroserviceDatabaseManager],
    stack: ExitStack,
    components_db_manager_cls: Type[MicroserviceDatabaseManager]) -> list[DockerContainerPorts]:
    """
    Starts the provided microservice containers.

    This function assumes that the testcontainers package is installed. Do not use it when testcontainers
    will not be available, for example, in any code that needs to run outside automated testing or the SwaggerUI
    interface.

    :param bridge_desktop_objects: data to go into the database that is internal to the docker network, as a map of
    collection -> list of desktop object dicts. This map must not be copied before being passed in.
    :param external_desktop_objects: data to go into the database that is outside of the docker network, as a map of
    collection -> list of desktop object dicts. This map must not be copied before being passed in.
    :param other_docker_images: a list of docker images to start.
    :param registry_url: the URL of the registry microservice.
    :param stack: the ExitStack.
    :param components_db_manager_cls: the database manager for registry components (required).
    """

    def _start_container_partial(img: DockerContainerConfig) -> DockerContainerPorts | None:
        if isinstance(img, MicroserviceContainerConfig):
            return _start_microservice(bridge_dbs, bridge_desktop_objects, external_desktop_objects, img, registry_url,
                                       stack,
                                       components_db_manager_cls)
        else:
            _start_other_container(img, stack)
            return None

    return [dcp for dcp in map(_start_container_partial, (i for i in other_docker_images or [])) if dcp is not None]


def _start_other_container(docker_container_config: DockerContainerConfig, stack: ExitStack) -> None:
    from .docker import start_other_container
    start_other_container(docker_container_config, stack)


def _start_microservice(bridge_dbs: list[MicroserviceDatabaseManager],
                        bridge_desktop_objects: MutableMapping[CollectionKey, MutableSequence[DesktopObjectDict]],
                        external_desktop_objects: MutableMapping[CollectionKey, MutableSequence[DesktopObjectDict]],
                        docker_container_config: MicroserviceContainerConfig, registry_url: Optional[str],
                        stack: ExitStack,
                        components_db_manager_cls: Type[MicroserviceDatabaseManager]) -> DockerContainerPorts:
    """
    Starts a docker container with the provided image. As a side effect, it adds HEA registry components to the
    bridge_desktop_objects and external_desktop_objects dictionaries.

    :return: a DockerContainerPorts instance.
    :raises ValueError: if the microservice failed to start.
    """
    logger = logging.getLogger(__name__)

    logger.debug('Bridge desktop objects before: %s', bridge_desktop_objects)
    logger.debug('External desktop objects before: %s', external_desktop_objects)
    from .docker import start_microservice_container
    db_manager = _get_db_manager(bridge_dbs, docker_container_config.db_manager_cls)
    if db_manager is not None:
        docker_container_config_ = docker_container_config.with_env_vars(db_manager.get_env_vars())
    else:
        docker_container_config_ = docker_container_config
    logger.debug('Starting container %s', docker_container_config_.image)
    try:
        external_url, bridge_url = start_microservice_container(docker_container_config_, stack, registry_url)
    except ValueError as e:
        raise ValueError(f'Failed to start docker container {docker_container_config_.image}') from e
    logger.debug('Container %s started', docker_container_config_.image)
    components_collection_key = CollectionKey(name='components', db_manager_cls=components_db_manager_cls)
    bridge_components = bridge_desktop_objects \
        .setdefault(
        get_collection_key_from_name(bridge_desktop_objects, name='components') or components_collection_key, [])
    items_to_add = docker_container_config_.get_bridge_components(bridge_url)
    bridge_components.extend(items_to_add)

    logger.debug('Bridge desktop objects after: %s', bridge_desktop_objects)
    external_desktop_objects \
        .setdefault(
        get_collection_key_from_name(external_desktop_objects, name='components') or components_collection_key, []) \
        .extend(docker_container_config_.get_external_components(external_url))
    logger.debug('External desktop objects after: %s', external_desktop_objects)
    return DockerContainerPorts(image=docker_container_config_.image, external_url=external_url, bridge_url=bridge_url)


def _generate_config_file(db_manager: MicroserviceDatabaseManager, registry_url: Optional[str]) -> str:
    """
    Generates a HEA microservice configuration file.

    :param db_manager: a DatabaseManager instance (required).
    :param registry_url: the URL of the registry service.
    :returns: the configuration file string.
    """
    if db_manager is not None:
        if registry_url is None:
            config_file = db_manager.get_config_file_section()
        else:
            config_file = f"""
    [DEFAULT]
    Registry={registry_url}

    {db_manager.get_config_file_section()}
                    """
    else:
        if registry_url is None:
            config_file = ''
        else:
            config_file = f"""
        [DEFAULT]
        Registry={registry_url}
                        """
    return config_file


def _with_hea_env_vars(container_config: MicroserviceContainerConfig,
                       registry_url: Optional[str]) -> MicroserviceContainerConfig:
    """
    Copies the provided container_spec, adding the environment variables corresponding to the provided arguments.

    This function assumes that the testcontainers package is installed. Do not use it when testcontainers
    will not be available, for example, in any code that needs to run outside automated testing or the SwaggerUI
    interface.

    :param container_config: the image and configuration (required).
    :param db_manager: a TestDatabaseFactory instance (required).
    :param registry_url: the URL of the registry service, which populates the HEASERVER_REGISTRY_URL environment
    variable.
    :return: the copy of the provided container_spec.
    """
    env_vars: Dict[str, str] = {}
    if registry_url is not None:
        env_vars['HEASERVER_REGISTRY_URL'] = registry_url
    return container_config.with_env_vars(env_vars)



import logging

from functools import wraps
from typing import Optional, Type, Any, TypeVar, Generic, AsyncContextManager
from collections.abc import Sequence, AsyncIterator, Awaitable, Coroutine, Callable, Generator, Iterable
from collections import OrderedDict
import requests
from contextlib import contextmanager
import asyncio
import time
import os
from dataclasses import dataclass
from heaobject.util import now  # Keep here so that older uses of it don't break.
import inspect
from contextlib import asynccontextmanager

RT = TypeVar('RT')  # return type

def async_retry(*exceptions: type[Exception], retries: int = 3, cooldown: int | None = 1,
                verbose = True) -> Callable[[Callable[..., Coroutine[Any, Any, RT]]], Callable[..., Coroutine[Any, Any, RT]]]:
    """
    Decorate an async function to execute it a few times before giving up.

    :param exceptions: One or more exceptions expected during function execution, as positional arguments.
    :param retries: Number of retries of function execution.
    :param cooldown: Seconds to wait before retry. If None, does not cool down.
    :param verbose: Specifies if we should log about not successful attempts.
    :raises: if all retries failed, or an exception other than those in the exceptions parameter was raised, the
    exception from the last try.
    """

    def wrap(func):
        @wraps(func)
        async def inner(*args, **kwargs) -> Any:
            _logger = logging.getLogger(__name__)
            retries_count = 0

            while True:
                try:
                    _logger.debug('About to try %s', func)
                    result = await func(*args, **kwargs)
                except exceptions as err:
                    message = "Exception {} during {} execution. " \
                              "{} of {} retries attempted".format(type(err), func, retries_count, retries)
                    if retries_count >= retries:
                        if verbose:
                            _logger.exception(message)
                        raise err
                    elif verbose:
                        _logger.warning(message)

                    if cooldown:
                        _logger.debug('Cooling down for %s seconds', cooldown)
                        await asyncio.sleep(cooldown)
                    retries_count += 1
                else:
                    return result
        return inner
    return wrap


def async_retry_context_manager(*exceptions: type[Exception], retries = 3, cooldown: Optional[int] = 1,
                                verbose = True) -> Callable[[Callable[..., AsyncContextManager[RT]]], Callable[..., AsyncContextManager[RT]]]:
    def wrap(ctxman: Callable[..., AsyncContextManager[RT]]) -> type[AsyncContextManager[RT]]:
        class AsyncContextManagerWrapper(AsyncContextManager[RT]):
            def __init__(self, *args, **kwargs) -> None:
                self.__args = args
                self.__kwargs = kwargs

            @async_retry(*exceptions, retries=retries, cooldown=cooldown, verbose=verbose)
            async def __aenter__(self) -> RT:
                self.__ctxman = ctxman(*self.__args, **self.__kwargs)
                return await self.__ctxman.__aenter__()

            async def __aexit__(self, exc_type, exc, tb) -> Any:
                return await self.__ctxman.__aexit__(exc_type, exc, tb)
        return AsyncContextManagerWrapper
    return wrap


def retry(*exceptions: Type[Exception], retries: int = 3, cooldown: Optional[int] = 1,
          verbose: bool = True) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """
    Decorate a non-async function to execute it a few times before giving up.

    :param exceptions: One or more exceptions expected during function execution, as positional arguments.
    :param retries: Number of retries of function execution.
    :param cooldown: Seconds to wait before retry. If None, does not cool down.
    :param verbose: Specifies if we should log about not successful attempts.
    :return a callable that accepts a callable, wraps it with the retry behavior, and returns a callable with the same
    signature.
    :raises: if all retries failed, or an exception other than those in the exceptions parameter was raised, the
    exception from the last try.
    """

    def wrap(func):
        @wraps(func)
        def inner(*args, **kwargs):
            _logger = logging.getLogger(__name__)
            retries_count = 0

            while True:
                try:
                    _logger.debug('About to try %s', func)
                    result = func(*args, **kwargs)
                except exceptions as err:
                    message = "Exception {} during {} execution. " \
                              "{} of {} retries attempted".format(type(err), func, retries_count, retries)

                    if retries_count >= retries:
                        verbose and _logger.exception(message)
                        raise err
                    else:
                        verbose and _logger.warning(message)

                    if cooldown:
                        _logger.debug('Cooling down for %s seconds', cooldown)
                        time.sleep(cooldown)
                    retries_count += 1
                else:
                    return result
        return inner
    return wrap


@contextmanager
def modified_environ(*remove, **update):
    """
    Temporarily updates the ``os.environ`` dictionary in-place.

    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations. Code obtained from
    https://stackoverflow.com/questions/2059482/python-temporarily-modify-the-current-processs-environment.

    :param remove: Environment variables to remove.
    :param update: Dictionary of environment variables and values to add/update.
    """
    env = os.environ
    update = update or {}
    remove = remove or []

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]


@contextmanager
def request_get_with_retry(url: str, retries=10, cooldown: int | None = 5, timeout: float | None = 60) -> Generator[requests.Response, None, None]:
    """
    Context manager that opens a URL, retrying if unsuccessful, and generates a HTTPResponse.

    :param url: the URL to open (required).
    :param retries: the number of retries.
    :param cooldown: the cooldown period (None for no cooldown).
    :param timeout: the connection timeout (None for no timeout).
    :raises TimeoutError: if the final retry timed out.
    :raises URLError: if the final retry errored out.
    """
    logger = logging.getLogger(__name__)
    logger.debug('Connecting to URL %s', url)

    if not retries:
        retries = 0

    @retry(requests.exceptions.RequestException, retries=retries, cooldown=cooldown)
    def do_open() -> requests.Response:
        try:
            with requests.get(url, timeout=timeout) as response:
                return response
        except Exception as e:
            logger.exception(f'Error getting URL {url}')
            raise e
    response = do_open()
    try:
        yield response
    finally:
        response.close()


class DuplicateError(ValueError):
    """
    The exception raised by the check_duplicates function.
    """
    def __init__(self, msg: str, duplicate_item: Any):
        """
        Creates the error.

        :param msg: the message to display when the exception is raised.
        :param duplicate_item: the item triggering the exception.
        """
        super().__init__(msg)
        self.__duplicate_item = duplicate_item

    @property
    def duplicate_item(self) -> Any:
        """The item triggering the exception."""
        return self.__duplicate_item


def check_duplicates(itr: Iterable[Any]):
    """
    Checks an iterable for duplicate items and raises an exception if one is found.

    @param itr: an iterable.
    @raise DuplicateError if a duplicate is found.
    """
    s = set()
    for item in itr:
        if item in s:
            raise DuplicateError(f'Duplicate {item}', item)
        else:
            s.add(item)


@dataclass(frozen=True)
class TypeEnforcingFrozenDataclass:
    def __post_init__(self):
        for (name, field_type) in self.__annotations__.items():
            if not isinstance(self.__dict__[name], field_type):
                current_type = type(self.__dict__[name])
                raise TypeError(f"The field `{name}` was assigned by `{current_type}` instead of `{field_type}`")

def issubclass_silent(cls, class_or_tuple, /):
    """
    Works like the builtin issubclass, except it returns False if cls is not a
    type instead of raising a TypeError.
    """
    return isinstance(cls, type) and cls is not None and issubclass(cls, class_or_tuple)

T = TypeVar('T')

def to_type(cls_or_instance: T | Type[T] | None) -> Type[T] | None:
    if cls_or_instance is None:
        return None
    elif isinstance(cls_or_instance, type):
        return cls_or_instance
    else:
        return type(cls_or_instance)

async def queued_processing(generator: Callable[[], AsyncIterator[T]],
                            item_processor: Callable[[T], Awaitable[None]],
                            exceptions_to_ignore: Callable[[Exception], bool] | None = None,
                            queue_max_size=100, num_workers=10):
    """
    Places items in an async generator into an asyncio queue for processing by the provided async item_processor.
    Specified exceptions may be suppressed.

    :param generator: the item async generator (required).
    :param item_processor: the async item processor.
    :param exceptions_to_ignore: a function that takes an exception parameter and returns whether the exception should
    be suppressed.
    :param queue_max_size the maximum size of the queue (default is 0, meaning infinite size).
    :param num_workers: the number of workers that take items off the queue (default is 10).
    """
    logger = logging.getLogger(__name__)
    queue: asyncio.Queue = asyncio.Queue(queue_max_size)
    async def worker(name: str, queue: asyncio.Queue):
        try:
            logger.debug('Starting worker %s', name)
            while True:
                coro = await queue.get()
                logger.debug('Worker %s processing item %s', name, coro)
                try:
                    await item_processor(coro)
                    logger.debug('Worker %s processing item %s', name, coro)
                finally:
                    queue.task_done()
        except asyncio.CancelledError:
            logger.debug('Ending worker %s', name)
        except:
            logger.exception('Error in worker %s', name)
            raise
    workers: list[asyncio.Task[None]] = []
    for i in range(num_workers):
        workers.append(asyncio.create_task(worker(f'worker-{i}', queue)))
    try:
        async for coro in generator():
            await queue.put(coro)
        await queue.join()
    finally:
        for worker_ in workers:
            worker_.cancel()
        exceptions_to_raise = []
        for r in await asyncio.gather(*workers, return_exceptions=True):
            if isinstance(r, Exception) and (not exceptions_to_ignore or not exceptions_to_ignore(r)):
                exceptions_to_raise.append(r)
        if exceptions_to_raise:
            raise exceptions_to_raise[0]


def public_methods(cls: Type) -> list[str]:
    """
    Returns a list of the provided class' public methods, defined as any callable attribute of the class that is not a
    class and whose name does not begin with an underscore.

    :param cls: the class (required).
    :return: a list of method names.
    """
    l: list[str] = []
    for method in dir(cls):
        attr = getattr(cls, method)
        if not method.startswith('_') and not inspect.isclass(attr):
            l.append(method)
    return l


async def yield_control():
    """
    Yields control from the current async task to another task.
    """
    await asyncio.sleep(0.1)


def raise_multiple(errors: Sequence[BaseException]):
    """
    Raises multiple exceptions at once, nesting them. The first exception in the sequence can be caught in a try-except
    clause around a call to this function.

    :param errors: the exceptions to raise.
    """
    if not errors:
        return
    errors_ = list(errors)
    try:
        raise errors_.pop()
    finally:
        raise_multiple(errors_)


K = TypeVar('K', bound=Any)

class AsyncioLockCache(Generic[K]):
    """
    A cache for asyncio.Lock objects with a fixed capacity, using an LRU eviction policy. K must be an immutable type.

    Attributes:
        capacity (int): The maximum number of locks the cache can hold.
        cache (OrderedDict[str, asyncio.Lock]): The internal cache storing the locks.
        ref_count (dict[str, int]): A dictionary tracking the reference count of each lock.
    """

    def __init__(self, capacity = 128) -> None:
        """
        Initialize the AsyncioLockCache with a given capacity.

        :param capacity: The maximum number of locks the cache can hold. Defaults to 128.
        """
        self.__capacity = capacity if capacity is not None else 128
        self.__cache: OrderedDict[K, asyncio.Lock] = OrderedDict()
        self.__ref_count: dict[K, int] = {}
        self.__lock = asyncio.Lock()

    @property
    def capacity(self) -> int:
        """
        The capacity of the cache (read-only).
        """
        return self.__capacity

    @property
    def cache(self) -> OrderedDict[K, asyncio.Lock]:
        """
        A copy of the internal cache (read-only).
        """
        return self.__cache.copy()

    @property
    def ref_count(self) -> dict[K, int]:
        """
        A copy of the reference count dictionary (read-only).
        """
        return self.__ref_count.copy()

    async def get(self, key: K) -> asyncio.Lock:
        """
        Get the lock associated with the given key, creating a new one if necessary.

        :param key: The key for which to get the lock.
        :return: The lock associated with the given key.
        """
        async with self.__lock:
            if key in self.__cache:
                self.__cache.move_to_end(key)
                self.__ref_count[key] += 1
            else:
                self.__cache[key] = asyncio.Lock()
                self.__ref_count[key] = 1
                if len(self.__cache) > self.__capacity:
                    await self.__evict()
            return self.__cache[key]

    async def release(self, key: K) -> None:
        """
        Release the lock associated with the given key, decrementing its reference count.

        :param key: The key for which to release the lock.
        """
        async with self.__lock:
            if key in self.__ref_count:
                self.__ref_count[key] -= 1
                if self.__ref_count[key] == 0:
                    del self.__ref_count[key]
                    del self.__cache[key]

    async def __evict(self) -> None:
        """
        Evict the least recently used lock that is not currently in use.
        """
        for key in list(self.__cache.keys()):
            if self.__ref_count[key] == 0:
                del self.__ref_count[key]
                del self.__cache[key]
                break


class LockManager(Generic[K]):
    """
    A context manager for acquiring and releasing asyncio locks for specific strings. K must be an immutable type.
    """

    def __init__(self, capacity = 128) -> None:
        """
        Initialize the LockManager.

        :param capacity:  The maximum number of locks the cache can hold. Defaults to 128.
        """
        self.__lock_cache: AsyncioLockCache[K] = AsyncioLockCache(capacity=capacity)

    @asynccontextmanager
    async def lock(self, key: K):
        """
        An async context manager for acquiring and releasing a lock for a specific key.

        :param key: The key for which to acquire the lock.
        """
        lock = await self.__lock_cache.get(key)
        try:
            async with lock:
                yield
        finally:
            await self.__lock_cache.release(key)



from unittest import TestCase, IsolatedAsyncioTestCase
from heaserver.service.db import awsservicelib
from heaobject.awss3key import encode_key
from botocore.exceptions import ClientError
from urllib.parse import quote

import heaserver.service.db.aws


class TestAWSServiceLib(TestCase):

    def test_decode_folder_root(self):
        self.assertEqual('', awsservicelib.decode_folder('root'))

    def test_decode_folder_non_root(self):
        self.assertEqual('/', awsservicelib.decode_folder('Lw=='))

    def test_decode_not_folder(self):
        self.assertEqual(None, awsservicelib.decode_folder('VGV4dEZpbGUucGRm'))

    def test_handle_client_error_404(self):
        c = ClientError(error_response={'Error': {'Code': heaserver.service.db.aws.CLIENT_ERROR_404}}, operation_name='foo')
        self.assertEqual(404, awsservicelib.handle_client_error(c).status)

    def test_handle_client_error_no_such_bucket(self):
        c = ClientError(error_response={'Error': {'Code': heaserver.service.db.aws.CLIENT_ERROR_NO_SUCH_BUCKET}}, operation_name='foo')
        self.assertEqual(404, awsservicelib.handle_client_error(c).status)

    def test_handle_client_error_unknown(self):
        c = ClientError(error_response={'Error': {'Code': "It's wrecked"}},
                        operation_name='foo')
        self.assertEqual(500, awsservicelib.handle_client_error(c).status)

class TestTestAWSServiceLib(IsolatedAsyncioTestCase):
    async def test_extract_source(self):
        actual = await awsservicelib._extract_source(
            {'bucket_id': 'arp-scale-2-cloud-bucket-with-tags11', 'id': encode_key('TestFolder2/')})
        self.assertEqual(('arp-scale-2-cloud-bucket-with-tags11', 'TestFolder2/'), actual)

    async def test_extract_target(self):
        actual = await awsservicelib._copy_object_extract_target({'template': {'data': [{'name': 'target',
                                                                                         'value': f'http://localhost:8080/volumes/12345678/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/{encode_key("TestFolder/")}'}]}})
        self.assertEqual(awsservicelib.TargetInfo(
                         'http://localhost:8080/volumes/12345678/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/VGVzdEZvbGRlci8=',
                         'arp-scale-2-cloud-bucket-with-tags11', 'TestFolder/', '12345678'), actual)

    async def test_extract_target_with_escaped_characters(self):
        actual = await awsservicelib._copy_object_extract_target({'template': {'data': [{'name': 'target',
                                                                                         'value': f'http://localhost:8080/volumes/12345678/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/{quote(encode_key("TestFolder/"))}'}]}})
        self.assertEqual(awsservicelib.TargetInfo(
                         f'http://localhost:8080/volumes/12345678/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/{quote(encode_key("TestFolder/"))}',
                         'arp-scale-2-cloud-bucket-with-tags11',
                         'TestFolder/', '12345678'),
            actual)





from aiohttp.test_utils import TestServer, TestClient, AioHTTPTestCase
from aiohttp.web import Application
from typing import Any, Union
from heaobject import root
from freezegun.api import FakeDatetime
# We need to override heaobject.root.json_encode because orjson only encodes datetimes not subclasses of datetime, and
# freezegun results in use of FakeDatetime objects.
def testing_json_encode(o: Any) -> Union[str, root.HEAObjectDict]:
    """
    Function to pass into the orjson.dumps default parameter that supports encoding HEAObjects and Fakedatetime objects.

    :param o: the object to encode.
    :return: the object after encoding.
    :raise TypeError: if the object is not a HEAObject.
    """
    match o:
        case root.HEAObject():
            return o.to_dict()
        case FakeDatetime():
            return o.isoformat()
        case _:
            raise TypeError(f'values {o} must be HEAObject, Fakedatetime, or a value type supported by orjson.dumps by default')
root.json_encode = testing_json_encode
from heaserver.service import appproperty
import pytest
import abc


class HEAAioHTTPTestCase(AioHTTPTestCase, abc.ABC):
    """
    Base class for testing HEA microservices.
    """

    def __init__(self, methodName=None, port=None):
        """
        Creates a test case object. The optional method name argument is passed into the superclass' constructor. If no
        port is specified in the constructor, a random port is selected. Test cases ignore the microservice's
        default port.

        :param methodName: the test method to execute.
        :param port: the port on which to run the service being tested.
        :raises ValueError: if the provided methodName does not exist.
        """
        super().__init__(methodName=methodName)
        self._port = port

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    @abc.abstractmethod
    async def get_application(self) -> Application:
        pass

    async def get_server(self, app: Application) -> TestServer:
        """
        Overrides this method to allow setting a fixed port for running aiohttp. If no port was specified in the
        constructor, a random port will be selected.

        :param app: the aiohttp application.
        :return: a new aiohttp TestServer instance.
        """
        if self._port:
            return TestServer(app, port=self._port)
        else:
            return TestServer(app)

    async def get_client(self, server: TestServer) -> TestClient:
        """Return a TestClient instance."""
        return TestClient(server, loop=self.loop, json_serialize=root.json_dumps)

    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.app[appproperty.HEA_CLIENT_SESSION] = self.client.session



from unittest import TestCase
from heaserver.service.jsonschemavalidator import WSTL_ACTION_SCHEMA_VALIDATOR, WSTL_SCHEMA_VALIDATOR, \
    CJ_TEMPLATE_SCHEMA_VALIDATOR, NVPJSON_SCHEMA_VALIDATOR, ValidationError


class WSTLActionSchemaValidatorTestCase(TestCase):
    def test_valid(self) -> None:
        """Checks if the WeSTL action schema validator does not raise an error when given valid JSON."""
        WSTL_ACTION_SCHEMA_VALIDATOR.validate({
            'name': 'openFile',
            'description': 'Open the selected file',
            'type': 'unsafe',
            'action': 'read',
            'target': 'foo bar',
            'prompt': 'Open',
            'rel': ['foo', 'bar']
        })

    def test_missing_required(self) -> None:
        """
        Checks if the WeSTL action schema validator raises ``ValidationError`` when given JSON that is missing a
        required property.
        """
        self.assertRaises(ValidationError, WSTL_ACTION_SCHEMA_VALIDATOR.validate, {
            'description': 'Open the selected file',
            'type': 'unsafe',
            'action': 'read',
            'target': 'foo bar',
            'prompt': 'Open',
            'rel': ['foo', 'bar']
        })

    def test_bad_type(self) -> None:
        """
        Checks if the WeSTL action schema validator raises ``ValidationError`` when given JSON with a property of an
        invalid type.
        """
        self.assertRaises(ValidationError, WSTL_ACTION_SCHEMA_VALIDATOR.validate, {
            'name': 'openFile',
            'description': 'Open the selected file',
            'type': 'unsafe',
            'action': 'read',
            'target': 'foo bar',
            'prompt': 'Open',
            'rel': ['foo', 'bar'],
            'inputs': 1
        })

    def test_empty(self) -> None:
        """
        Checks if the WeSTL action schema validator raises ``ValidationError`` when given an empty dictionary (``{}``).
        """
        self.assertRaises(ValidationError, WSTL_ACTION_SCHEMA_VALIDATOR.validate, {})


class WSTLSchemaValidatorTestCase(TestCase):
    def test_valid(self) -> None:
        """Checks if the WeSTL schema validator does not raise an error when given valid JSON."""
        WSTL_SCHEMA_VALIDATOR.validate({
            'wstl': {
                'title': 'Foo Bar'
            }
        })

    def test_missing_required(self) -> None:
        """
        Checks if the WeSTL schema validator raises ``ValidationError`` when given JSON that is missing the "wstl"
        property.
        """
        self.assertRaises(ValidationError, WSTL_SCHEMA_VALIDATOR.validate, {
            'notWstl': {
                'title': 'Foo Bar'
            }
        })

    def test_bad_type(self) -> None:
        """
        Checks if the WeSTL schema validator raises ``ValidationError`` when given JSON with a property of an
        invalid type.
        """
        self.assertRaises(ValidationError, WSTL_SCHEMA_VALIDATOR.validate, {
            'wstl': {
                'title': 'Foo Bar',
                'data': [{
                    'subdata': {
                        'foo': 'bar'
                    }
                }]
            }
        })

    def test_empty(self) -> None:
        """
        Checks if the WeSTL schema validator raises ``ValidationError`` when given an empty dictionary (``{}``).
        """
        self.assertRaises(ValidationError, WSTL_SCHEMA_VALIDATOR.validate, {})


class CJTemplateSchemaValidatorTestCase(TestCase):
    def test_valid(self) -> None:
        """Checks if the Collection+JSON template schema validator does not raise an error when given valid JSON."""
        CJ_TEMPLATE_SCHEMA_VALIDATOR.validate({
            'template': {
                'data': [{
                    'name': 'foo',
                    'value': 'bar'
                }]
            }
        })

    def test_missing_required_data_property(self) -> None:
        """
        Checks if the Collection+JSON template schema validator raises ``ValidationError`` when given Collection+JSON
        with data that is missing the "name" property.
        """
        self.assertRaises(ValidationError, CJ_TEMPLATE_SCHEMA_VALIDATOR.validate, {
            'template': {
                'data': [{
                    'value': 'bar'
                }]
            }
        })

    def test_bad_type(self) -> None:
        """
        Checks if the Collection+JSON template schema validator raises ``ValidationError`` when given JSON with a
        property of an invalid type.
        """
        self.assertRaises(ValidationError, CJ_TEMPLATE_SCHEMA_VALIDATOR.validate, {
            'template': {
                'data': 'The quick brown fox jumps over the lazy dog'
            }
        })

    def test_empty(self) -> None:
        """
        Checks if the Collection+JSON template schema validator raises ``ValidationError`` when given an empty
        dictionary (``{}``).
        """
        self.assertRaises(ValidationError, CJ_TEMPLATE_SCHEMA_VALIDATOR.validate, {})


class NVPJSONSchemaValidatorTestCase(TestCase):
    def test_valid(self) -> None:
        """Checks if the name-value-pair JSON schema validator does not raise an error when given valid JSON."""
        NVPJSON_SCHEMA_VALIDATOR.validate({
            'foo': 'bar'
        })

    def test_not_json(self) -> None:
        """
        Checks if the name-value-pair JSON schema validator raises ``ValidationError`` when not given a dictionary.
        """
        self.assertRaises(ValidationError, NVPJSON_SCHEMA_VALIDATOR.validate, 'foo')



import logging

from aiohttp import web
from aiohttp.web_request import Request
from aiohttp.web_response import Response
from heaobject.user import NONE_USER
from opensearchpy import AsyncOpenSearch
import re

from heaserver.service import response
from heaserver.service.appproperty import HEA_DB
from heaserver.service.db.database import DatabaseContextManager
from heaserver.service.db.opensearch import OpenSearch, ItemTypeVar, OpenSearchContext
from heaserver.service.oidcclaimhdrs import SUB


def build_query(request: Request, permission_context: dict[str, list[str]]):
    # Placeholder for the must array in the query
    # Handle multiple `regexp` query params
    search_term = request.query.get('text', '')
    santize_search_term = re.escape(search_term)
    regexp_type = request.query.get('regexp_type', 'contains')

    if not search_term:
        raise ValueError("search text is required and cannot be empty")

    # Build the final query object
    if regexp_type == 'contains':
        regex_value = f".*{santize_search_term}.*"
    elif regexp_type == 'starts_with':
        regex_value = f"{santize_search_term}.*"
    elif regexp_type == 'ends_with':
        regex_value = f".*{santize_search_term}"
    else:
        raise ValueError("Invalid query type. Must be one of: 'contains', 'starts_with', 'ends_with'.")

    query = {
        "query": {
            "bool": {
                "must": [
                    {
                        "regexp": {
                            "path": {
                                "value": regex_value
                            }
                        }
                    },
                    {"terms": permission_context}
                ]
            }
        }
    }

    return query


async def search(request: Request,
                 search_item_type: type[ItemTypeVar],
                 perm_context: dict[str, list[str]],
                 index: str | None = None,
                 volume_id: str | None = None) -> Response:
    """
    Executes a dynamic search query in OpenSearch and returns the result.

    :param request: the HTTP request.
    :param search_item_type:
    :param perm_context: The permission context provides list of strings where at least one needs match
    :param index: (Optional) the OpenSearch index name, if not present it will default to db config for it.
    :param volume_id: (Optional) the id of the volume
    :return: The OpenSearch query result or None if no result is found.
    """
    sub = request.headers.get('SUB', 'none_user')
    logger = logging.getLogger(__name__)
    volume_id_ = request.match_info.get('volume_id') if request.match_info.get('volume_id', None) else volume_id

    # Build the query dynamically based on the input parameters
    query = build_query(request, perm_context)
    async with OpenSearchContext(request=request, volume_id=volume_id_) as opensearch:
        results = await opensearch.search(query=query, search_item_type=search_item_type, index=index)
        if not results:
            return response.status_bad_request("Invalid search results")
        logger.debug(f"opensearch result: {results}")
    return await response.get_all(request, [sr.to_dict() for sr in results])

